#include "msnobject.h"
#include "msn.h"

/*
 * BASE64 encoding, functions comes from yahoo protocol module
 * just some little mods..
 */
static char msn_base64digits[] ="ABCDEFGHIJKLMNOPQRSTUVWXYZ"
                                "abcdefghijklmnopqrstuvwxyz"
                                "0123456789+/=";
void msn_tobase64( unsigned char *out, const unsigned char *in, int inlen )
{
        for ( ; inlen >= 3; inlen -= 3 )
                {
                        *out++ = msn_base64digits[in[0] >> 2];
                        *out++ = msn_base64digits[( ( in[0]<<4 ) & 0x30 ) | ( in[1]>>4 )];
                        *out++ = msn_base64digits[( ( in[1]<<2 ) & 0x3c ) | ( in[2]>>6 )];
                        *out++ = msn_base64digits[in[2] & 0x3f];
                        in += 3;
                }
        if ( inlen > 0 )
                {
                        unsigned char fragment;

                        *out++ = msn_base64digits[in[0] >> 2];
                        fragment = ( in[0] << 4 ) & 0x30;
                        if ( inlen > 1 )
                                fragment |= in[1] >> 4;
                        *out++ = msn_base64digits[fragment];
                        *out++ = ( inlen < 2 ) ? '='
                                        : msn_base64digits[( in[1] << 2 ) & 0x3c];
                        *out++ = '=';
                }
        *out = '\0';
}

/*
 * convert an http-encoded msnobject to an msn context
 * ( = base64(<http-decoded msnobject>+<\x00>) )
 */
char *msnobject_to_context( char *msnobject )
{
	char *buf1, *buf2;
	int length = strlen( msnobject );
	
	buf1 = g_new0( char, length + 2 );
	if ( !buf1 )
		return NULL;

	strncpy( buf1, msnobject, length );
	http_decode( buf1 );
	
	length = strlen( buf1 );

	buf2 = g_new0( char, length * 3 );

	if ( !buf2 )
	{
		g_free( buf1 );
		return NULL;
	}

	msn_tobase64( buf2, buf1, length + 1 );

	g_free( buf1 );
	
	return buf2;
}

/*
 * create a new msnobject from the specified parameters
 */
char *msnobject_from_params( char *creator, int type, char *filename )
{
	SHA_CTX SHA_Context;
	FILE *fd;
	struct stat statinfo;
	unsigned char digest[20];

	char buffer[1024], sha1d[200], sha1c[200], *tmp, *location;
	int len;

	if ( !filename )
		return NULL;
	
	if ( ( stat( filename, &statinfo ) ) != 0 )
		return NULL;
		
	if ( !( fd=fopen( filename, "r" ) ) )
		return NULL;
	
	tmp = g_strdup( filename );
	if ( !tmp )
		return NULL;
	location = basename( tmp );
	g_free( tmp );

	/* calculate SHA1D */
	shaInit( &SHA_Context );
	while ( ( len = fread( buffer, 1, sizeof( buffer ), fd ) ) > 0 )
	{
		shaUpdate( &SHA_Context, buffer, len );
	}
	shaFinal( &SHA_Context, digest );
	msn_tobase64( sha1d, digest, 20 );
	fclose( fd );
	http_encode( sha1d );

	/* calculate SHA1C */
	shaInit( &SHA_Context );
	len = g_snprintf( buffer, sizeof( buffer ),
		"Creator%sSize%dType%dLocation%sFriendlyAAA=SHA1D%s",
		creator, ( int )statinfo.st_size, type, location, sha1d );
	shaUpdate( &SHA_Context, buffer, len );
	shaFinal( &SHA_Context, digest );
	msn_tobase64( sha1c, digest, 20 );
	http_encode( sha1c );

	/* create msnobject */
	g_snprintf( buffer, sizeof( buffer ),
		"%%3Cmsnobj%%20Creator%%3D%%22%s%%22%%20Size%%3D%%22%d%%22%%20Type%%3D%%22%d%%22%%20Location%%3D%%22%s%%22%%20Friendly%%3D%%22AAA%%3D%%22%%20SHA1D%%3D%%22%s%%22%%20SHA1C%%3D%%22%s%%22/%%3E",
		creator, ( int )statinfo.st_size, type, location, sha1d, sha1c );

	return g_strdup( buffer );
}

/*
 * returns a field from an msnobject..
 */
char *msnobject_get_field( char *field, char *msnobject )
{
        char *tag = NULL, *start = NULL, *end = NULL, tmp[200];

        char *decoded_msnobject = g_new0( char, strlen( msnobject ) * 3 );

        if ( !decoded_msnobject )
                return NULL;

        strcpy( decoded_msnobject, msnobject );
        http_decode( decoded_msnobject );

        g_snprintf( tmp, sizeof( tmp ), "%s=\"", field );

        if ( ( start = strstr( decoded_msnobject, tmp ) ) != NULL )
        {
                start += strlen( tmp );
                end = strchr( start, '"' );
                if ( end )
                        tag = g_strndup( start, end-start );
        }

        g_free( decoded_msnobject );

        return tag;
}

